(in-package :cl-user)

(defpackage :if-lib
  (:use :common-lisp :if-basic-lib)
  (:export :container :room :item :clothing :capacity
	   :n-to :ne-to :e-to :se-to :s-to :sw-to :w-to :nw-to :in-to :out-to
	   :*intscope*  :*outscope* :*location* :*trace-light* :*vowels* 
	   :*score* :*gamestate* :*turns*
	   :compass :dir-n :dir-ne :dir-e :dir-se :dir-s
	   :dir-sw :dir-w :dir-nw :dir-u :dir-d :dir-in :dir-out
	   :darkness :lit :transparent :passable
	   :reachp :seep :global-reachp :global-seep :seep1 :reachp1
	   :darkness :actor :selfobj :*player* :find-location
	   :deduce-article :print-property :print-name :list-contents
	   :reset-scope :look :words2dic-first :disambig :normal-token-scope
	   :input-quit-loop :quit-game :prompt :prompt-read :input-loop
	   :go-to-room :property :init :test-seq
	   :heldp :the-name
	   :supporter :animate
	   :afterlife :print-gamestate :end-game
	   )
  (:shadow :room))

(in-package :if-lib)

(defparameter *vowels* "aeiouy")

;;Access to shadowed room function

(defun room (&optional (arg :default))
  (cl:room arg))

(define-compiler-macro room (&whole whole &optional arg)
  (declare (ignore arg))
  `(cl:room ,@(cdr whole)))

;;---------------

(ifclass container () (capacity integer) (has :container))
(ifclass supporter () (capacity integer) (has :supporter))


(ifclass room () (description string)
	 (n-to object) (ne-to object) (e-to object) (se-to object)
	 (s-to object) (sw-to object) (w-to object) (nw-to object)
	 (u-to object) (d-to object) (in-to object) (out-to object)
	 (has :light :enterable))

(ifclass item () (description string) (glance string) (article string)
	 (has :item))

(ifclass clothing (item) (has :clothing))

(object darkness (room) 
	(description "It's pitch black. You can't see a thing")
	(has :~light))

;;Compass directions
(object compass ())
(object dir-n () "north" (name "north" "n") compass (property 'n-to))
(object dir-ne () "northeast" (name "northeast" "ne") compass 
	(property 'ne-to))
(object dir-e () "east" (name "east" "e") compass (property 'e-to))
(object dir-se () "southeast" (name "southeast" "se") compass 
	(property 'se-to))
(object dir-s () "south" (name "south" "s") compass (property 's-to))
(object dir-sw () "southwest" (name "southwest" "sw") compass 
	(property 'sw-to))
(object dir-w () "west" (name "west" "w") compass (property 'w-to))
(object dir-nw () "northwest" (name "northwest" "nw") compass 
	(property 'nw-to))
(object dir-u () "up" (name "up" "u") compass (property 'u-to))
(object dir-d () "down" (name "down" "d") compass (property 'd-to))
(object dir-in () "in"  compass (property 'in-to))
(object dir-out () "out" compass (property 'out-to))

(defparameter *intscope* nil)
(defparameter *outscope* nil)
(defparameter *location* nil)
(defparameter *score* 0)
(defparameter *gamestate* 0)
(defparameter *turns* 0)

(ifclass actor () (reachp function nil) (seep function nil))
(ifclass animate () (has :animate))

(object selfobj (actor animate) "me" 
	(article "")
	(has :scenery))

(defparameter *player* selfobj)

(defun find-location (obj)
  (loop for o = obj then (parent o) while (parent o) finally (return o))) 



(defun transparent (obj)
  (or (has obj :container :open) 
      (has obj :supporter)
      (has obj :transparent)
      (eql obj *player*)))

(defun lit-down (obj)
  (if (has obj :light) t
      (some #'(lambda (x) (or (has x :light) 
			      (and (transparent x) (lit-down x))))
	    (children obj))))

(defun lit (obj)
  (or (lit-down obj) 
      (when (parent obj) 
	(or (has (parent obj) :light) 
	    (and (transparent (parent obj)) (lit (parent obj)))))))


(defun seep-down (actor obj)
  (if (eql obj actor) t
      (some #'(lambda (x) (or (eql x actor) 
			      (and (transparent x) (seep-down actor x))))
	    (children obj))))
 

(defun seep2 (actor obj)
  (or (seep-down actor obj) 
      (when (parent obj) 
	(or (eql (parent obj) actor) 
	    (and (transparent (parent obj)) (seep2 actor (parent obj)))))))

(defun seep1 (actor obj)
  (and (lit actor) (lit obj) (or (in obj compass) (seep2 actor obj))))

(defun global-seep (actor obj)
  "Tests whether an object is seen by actor"
  (if (and (typep actor 'actor) (seep actor))
      (read-property actor 'seep obj)
      (seep1 actor obj)))

(defun passable (obj)
  (or (has obj :container :open)
      (has obj :supporter)
      (eql obj *player*)))


(defun reachp-down (actor obj)
  (if (eql obj actor) t
      (some #'(lambda (x) (or (eql x actor) 
			      (and (passable x) (reachp-down actor x))))
	    (children obj))))
 

(defun reachp2 (actor obj)
  (or (reachp-down actor obj) 
      (when (parent obj) 
	(or (eql (parent obj) actor) 
	    (and (passable (parent obj)) (reachp2 actor (parent obj)))))))


(defun reachp1 (actor obj)
  (and (lit obj) (lit actor) (reachp2 actor obj)))
  
(defun global-reachp (actor obj)
  "Tests whether an object is reachable by actor"
  (if (and (typep actor 'actor) (reachp actor))
      (read-property actor 'reachp obj)
      (reachp1 actor obj)))


(defun deduce-article (name)
  (let ((firstchar (aref name 0)))
    (cond ((char= firstchar (char-upcase firstchar)) "the")
	  ((find firstchar *vowels* :test #'char=) "an")
	  (t "a")))) 

(defun print-property (obj property)
  (multiple-value-bind (value printp) (read-property obj property)
    (if (and (stringp value) (not printp)) (sprint value) value)))

(defgeneric print-name (obj &key article capital)
  (:documentation "Returns a string containing the name of object"))

(defgeneric the-name (obj &key article capital)
  (:documentation "Returns a string containing _the_ name of object"))

(defmethod print-name ((obj abstractobject) &key (article nil) (capital nil))
  "Prints a name with article"
  (let ((outstr
  (with-output-to-string (out)
    (let ((*standard-output* out)
	  (name (read-property obj 'name)))
      (if article
	  (unless (zerop (length article)) (format t "~a " article))
      (if (provides obj article) 
	  (print-property obj 'article)
	  (progn (princ (deduce-article name)) (princ " "))))
      (princ name)))))
    (when capital (setf (aref outstr 0) (char-upcase (aref outstr 0))))
    outstr))

(defmethod print-name ((obj clothing) &key &allow-other-keys)
  (if (has obj :worn) (concatenate 'string (call-next-method) " (worn)")
      (call-next-method)))

(defmethod print-name ((obj item) &key &allow-other-keys)
  (if (has obj :light) (concatenate 'string (call-next-method) 
				    " (providing light)")
      (call-next-method)))

(defmethod print-name ((obj animate) &key (article nil) (capital nil))  
  (call-next-method obj :article (or article "") :capital capital))

(defmethod the-name ((obj abstractobject) &key (article nil) (capital nil))
  (let ((outstr
  (with-output-to-string (out)
    (let ((*standard-output* out)
	  (name (read-property obj 'name)))
      (if article
	  (unless (zerop (length article)) (format t "~a " article))
      (if (provides obj article) 
	  (print-property obj 'article)
	  (progn (princ "the "))))
      (princ name)))))
    (when capital (setf (aref outstr 0) (char-upcase (aref outstr 0))))
    outstr))
	
(defmethod the-name ((obj animate) &key (article nil) (capital nil))  
  (call-next-method obj :article (or article "") :capital capital))


(defun print-inside (obj stream)
  (let ((*standard-output* stream))
  (when (has obj :container)
    (if (or (has obj :open) (has obj :transparent))
	(when (children obj)
	  (progn (princ " (containing ") 
		 (princ (list-contents obj)) 
		 (princ ")"))
	  (princ " (empty)"))
	(princ " (closed)")))
  (when (has obj :supporter)
    (when (children obj)
      (progn (princ " (on top of which are ") 
	     (princ (list-contents obj)) 
	     (princ ")"))))))

(defun list-contents (obj)
  (with-output-to-string (out)
  (let (commaflag)
    (when (children obj) 
      (objectloop (in x obj)
		  (if (and !last! commaflag) (princ " and " out)
		      (when commaflag (princ ", " out)))
		  (setf commaflag t)
		  (princ (print-name x) out)
		  (print-inside obj out))))))
	

(defun default-glance (obj)
  (sprint "~a" (with-output-to-string (out)
  (let ((*standard-output* out))
    (princ "There is ") (princ (print-name obj)) 
    (print-inside obj out)
    (princ ".")))))


(defun reset-scope (&optional (location :unspecified))
  (if (eql location :unspecified) (setf location *location*))
  (unless location (setf *intscope* nil *outscope* nil) 
	  (return-from reset-scope))
  (setf *outscope* (loop for x in *allobjects* ;unless (eql x location)
		      when (global-seep location x) collect x))
  (setf *intscope* (loop for x in *allobjects* ;unless (eql x location)
		      when (global-reachp location x) collect x))
  (let ((int (loop for x in *intscope* 
		   when (provides x 'in-scope) 
		   append (read-property x 'in-scope)))
	(out (loop for x in *outscope* 
		   when (provides x 'in-scope) 
		   append (read-property x 'in-scope))))
    (setf *outscope* (nconc *outscope* out))
    (setf *intscope* (nconc *intscope* int))))
      
(defgeneric look (obj) 
  (:documentation "Used for looking in rooms and containers"))

(defmethod look ((room room))
  (sprint "~a~%~%" (read-property room 'name))
  (when (provides room description) (print-property room 'description))
  (freshline)
  (objectloop (in x room)
	      (when (hasnt x :scenery :hidden)
	      (freshline)
	      (if (provides x glance)
		  (if (print-property x 'glance) (default-glance x))
		  (default-glance x)))))

;;Tokens

(defun greedy-match-noun (obj words)
  "Finds how many words obj matches"
  (loop for w in words
	collecting w into ww
	when (< (parser obj ww) 1) return (length (butlast ww))
	finally (return (length words))))
	

(defun find-best-match (words scope)
  (loop with bestvalue = 0
	with bestnouns = nil
        for x in scope
	for y = (greedy-match-noun x words)
	when (> y bestvalue) do (setf bestvalue y) (setf bestnouns (list x))
	else when (and (= y bestvalue) (> bestvalue 0)) do (push x bestnouns)
	finally (return (values bestnouns bestvalue))))

(defun words2dic-first (words)
  (loop for w in words
	for n = (handler-case (word2dic w) (nosuchword () nil))
	while n collect n))

(define-condition disambig ()
  ((what :initarg :what :initform nil :reader disambig-what))
  (:report (lambda (condition stream) 
	     (format stream "Cannot decide between the objects: ~A" 
		     (disambig-what condition)))))  

(defun normal-token-scope (wordlist scope)
  (multiple-value-bind (bnouns bvalue) 
	  (find-best-match (words2dic-first wordlist) scope)
	(case (length bnouns)
	  (1 (values t (car bnouns) 
		     (last wordlist (- (length wordlist) bvalue))))
	  (0 nil)
	  (t (let ((choosewhat (signal 'disambig :what bnouns)))
	       (when choosewhat 
		 (values t choosewhat
			 (last wordlist (- (length wordlist) bvalue)))))))))

;;Token definitions

(deftoken :noun ;matches the given scope (intscope by default)
    (let ((scope (if (eql *tokenpar* :unspecified) *intscope* *tokenpar*))) 
      (normal-token-scope wordlist scope)))

(deftoken :seen ;matches outscope
    (normal-token-scope wordlist *outscope*))
					 
(deftoken :or  ;An ugly hack - matches either of supported words
    (if (member (car wordlist) *tokenpar* :test #'string==)
	(values t nil (cdr wordlist)) nil))

(deftoken :has ;has flag, in outscope
    (normal-token-scope wordlist 
        (remove-if #'(lambda (x) (hasnt x *tokenpar*)) *outscope*)))

(deftoken :direction ;compass direction
    (normal-token-scope wordlist (children compass)))

(defun heldp (obj)
  (if (eql obj *player*) t
      (and (parent obj) (transparent (parent obj)) (heldp (parent obj)))))

(deftoken :held
    (normal-token-scope wordlist
	(remove-if (complement #'heldp) *outscope*)))	       
  
;;Action helpers

(defun go-to-room (room)
  (when (typep room 'abstractobject)
    (setf *location* room)
    (move *player* *location*)
    (reset-scope *location*)
    (look *location*)))
    
;;Main loop
(define-condition input-quit-loop () ()) 

(defun quit-game ()
  (signal 'input-quit-loop))

(defun prompt ()
  (princ "> " *outstream*)) 

(defun prompt-read ()
  (newline *outstream*) (prompt)
  (force-output *outstream*)
  (read-line *instream*))

(defun input-loop ()
  (handler-case 
      (loop (unless (zerop *gamestate*) (end-game))
	    (parse-command (prompt-read)))
      (input-quit-loop () nil)))

;(defun init () "Warning: No init routine detected!") 

(declaim (ftype function init))

(defun afterlife () "This is a stub" t)

;(defun print-gamestate () "You have lost")
(declaim (ftype function print-gamestate))

(defun print-gamestate-default ()
  (case *gamestate*
    (1 "You have died")
    (2 "You have won")
    (t (print-gamestate))))

(defun end-game ()
  (when (afterlife)
    (sprint "~%~%~%***~a***~%~%~%" (print-gamestate-default))
    (sprint "Score:~a Turns:~a" *score* *turns*)
    (quit-game)))


(defun test-seq ()
  (setf *score* 0 *turns* 0)
  (exec init ())
  (go-to-room *location*)
  (input-loop))

(defun turn-passing (&optional time)
  (reset-scope)
  (incf *turns* time))

;;Simpler defpackage for easy using

(defmacro load-libs (name &rest other-packages)
  `(defpackage ,name
    (:use :common-lisp :if-basic-lib :if-lib :verb-lib ,@other-packages)
    (:shadowing-import-from :if-lib :room)
    ))
   

