(in-package :cl-user)

(defpackage :verb-lib
  (:use :common-lisp :if-lib :if-basic-lib)
  (:export :attack :take :teleport :examine :look-around
	   :go-to :go-n :go-ne :go-e :go-se :go-s :go-sw :go-w :go-nw
	   :go-u :go-d :go-in :go-out
	   :take :put-in :put-on :drop :receive)
  (:shadowing-import-from :if-lib :room))

(in-package :verb-lib)

(verb "quit" '(-> quit-game)) ;;That one you'll use often ;)

;Debug verb
(verb "teleport"
      `((:noun ,(lambda () *allobjects*)) -> teleport))

(verb "take" "get"
      '(:noun -> take)) 
   
(verb "look"
      '(-> look-around)
      '("at" :seen -> examine))

(verb "examine" "x"
      '(:noun -> examine))

(verb "attack" "break" "crack" "destroy"
     "fight" "hit" "kill" "murder" "punch"
     "smash" "thump" "torture" "wreck"
    '(:noun -> attack))

(verb "go" "run" "walk" '(:direction -> go-to))
(verb "n" "north" '(-> go-n))
(verb "ne" "northeast" '(-> go-ne))
(verb "e" "east" '(-> go-e))
(verb "se" "southeast" '(-> go-se))
(verb "s" "south" '(-> go-s))
(verb "sw" "southwest" '(-> go-sw))
(verb "w" "west" '(-> go-w))
(verb "nw" "northwest" '(-> go-nw))
(verb "up" '(-> go-up))
(verb "down" '(-> go-down))
(verb "in" '(-> go-in))
(verb "out" '(-> go-out))

(verb "inventory" "i" '(-> inventory))

(verb "take" "get"
      '(:noun -> take))

(verb "put" "drop"
      '(:held "on" :noun -> put-on)
      '(:held "in" :noun -> put-in)
      '(:held "down" -> drop)
      '(:held -> drop))

(defaction attack (obj) "Violence is not the answer.")

(defaction teleport (obj) 
  (go-to-room obj))

(defaction examine (obj) (print-property obj 'description) 
	   (terpri *outstream*)) 

(defun look-around () (look *location*))

(defaction go-to (dir)
  (go-to-room (read-property *location* (property dir))))

(defaction go-n () (run-action 'go-to dir-n))
(defaction go-ne () (run-action 'go-to dir-ne))
(defaction go-e () (run-action 'go-to dir-e))
(defaction go-se () (run-action 'go-to dir-se))
(defaction go-s () (run-action 'go-to dir-s))
(defaction go-sw () (run-action 'go-to dir-sw))
(defaction go-w () (run-action 'go-to dir-w))
(defaction go-nw () (run-action 'go-to dir-nw))
(defaction go-u () (run-action 'go-to dir-u))
(defaction go-d () (run-action 'go-to dir-d))
(defaction go-in () (run-action 'go-to dir-in))
(defaction go-out () (run-action 'go-to dir-out))

(defun inventory ()
  (sprint "You are carrying: ~a." (list-contents *player*))
  (newline))

(defaction take (obj)
  "You can't take that")

(defmethod take((obj item))
  (if (has obj :item)
      (progn
	(move obj *player*)
	(when (run-action-after obj) "Taken."))
      (call-next-method)))

(defaction drop (obj)
  (unless (has obj :item) (return-from drop "You can't drop that."))
  (move obj (parent *player*))
  (give obj :~worn)
  (when (run-action-after obj) "Dropped."))

(defaction put-on (item host)
  "You can't put anything on that.")

(defmethod put-on ((item item) (host supporter))
  ;;(format t "(~a ~a)" (print-name item) (print-name host)) 
  (unless (has item :item) (return-from put-on "You can't get rid of that."))
  (unless (has host :supporter) (return-from put-on (call-next-method)))
  (and (run-action 'receive (reverse *args*) :time 0)
       *after*
       (run-action-after item) 
       "Done."))

(defaction put-in (item host)
  "You can't put anything in that.")

(defmethod put-in ((item item) (host container))
  (unless (has item :item) (return-from put-in "You can't get rid of that."))
  (unless (has host :container) (return-from put-in (call-next-method)))
  (when (has host :closed) 
    (return-from put-in 
      (format nil "~a is closed." (the-name host :capital t))))
  (and (run-action 'receive (reverse *args*) :time 0)
       *after*
       (run-action-after item) 
       "Done."))
    
(defaction receive (host guest)
  "No method defined for that kind of object movement.")

(defmethod receive ((host supporter) (item item))
  (if (or (zerop (capacity host)) 
	  (< (list-length (children host)) (capacity host)))
    (progn (move item host)
	   (run-action-after host))
    "Not enough space."))

(defmethod receive ((host container) (item item))
  (if (or (zerop (capacity host)) 
	  (< (list-length (children host)) (capacity host)))
    (progn (move item host)
	   (run-action-after host))
    "Not enough space."))